""":mod:`kamis.models.common` --- KAMIS 오픈 API 공통 모델 모듈
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KAMIS 오픈 API의 요청에 필요한 공통된 모델들을 정의한 모듈입니다.

"""
from dataclasses import dataclass
import enum
from typing import Dict, List, Union


__all__ = 'Country', 'Direction', 'BasicResponse', 'ProductClass', 'ResultCode'


class Country(enum.Enum):
    """지역"""
    SEOUL = '1101'
    BUSAN = '2100'
    DAEGU = '2200'
    INCHEON = '2300'
    GWANGJU = '2401'
    DAEJEON = '2501'
    ULSAN = '2601'
    SUWON = '3111'
    CHUNCHEON = '3211'
    CHEONGJU = '3311'
    JEONJU = '3511'
    POHANG = '3711'
    JEJU = '3911'
    UIJEONGBU = '3113'
    SUNCHEON = '3613'
    ANDONG = '3714'
    CHANGWON = '3814'
    YONGIN = '3145'


class Direction(enum.Enum):
    """등락여부"""
    #: 가격하락
    DECREASE = '0'
    #: 가격상승
    INCREASE = '1'
    #: 등락없음
    NO_CHANGE = '2'


class ProductClass(enum.Enum):
    """품목구분"""
    #: 소매
    RETAIL = '01'
    #: 도매
    WHOLESALE = '02'


class ResultCode(enum.Enum):
    """비고/에러 메세지"""
    #: 성공
    SUCCESS = '000'
    #: 결과 없음
    NO_DATA = '001'
    #: 잘못된 파라미터
    WRONG_PARAMETERS = '200'
    #: 인증되지 않은 요청
    UNAUTHENTICATED_REQUEST = '900'

    @property
    def succeeded(self):
        return self in (ResultCode.SUCCESS, ResultCode.NO_DATA)


@dataclass(frozen=True)
class BasicResponse:
    """모든 요청의 응답에 공통적으로 포함되어 있는 값을 갖는 클래스입니다."""
    #: API 문서의 "요청 메세지"
    condition: Union[List[List[str]], Dict[str, str]]



""":mod:`kamis.client` --- KAMIS 오픈 API 클라이언트 모듈
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KAMIS 오픈 API를 실행하기 위한 클래스 :class:`KamisOpenApi`를 정의한
모듈입니다.

"""
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Union
import urllib.parse

from requests import get

from .models.common import ResultCode
from .models.daily_sales_list import (
    PriceResponse as DailySalesListPriceResponse,
    Response as DailySalesListResponse,
)
from .models.period_product_list import (
    DataResponse as PeriodProductListDataResponse,
    ItemResponse as PeriodProductListItemResponse,
    Parameters as PeriodProductListParameters,
    Response as PeriodProductListResponse,
)


__all__ = 'CertificationPair', 'KamisOpenApi'


@dataclass
class CertificationPair:
    """인증 정보를 담는 클래스입니다."""
    #: 인증 키
    cert_key: str
    #: 요청자 아이디
    cert_id: str


class KamisOpenApi:
    """오픈 API를 실행할 수 있는 클라이언트 클래스입니다."""
    def __init__(
        self,
        cert: CertificationPair,
        endpoint: str = "http://www.kamis.or.kr/service/price/xml.do"
    ):
        self.cert = cert
        self.endpoint = endpoint

    def build_action_endpoint(
        self,
        action: str,
        parameters: Dict[str, Union[str, List[str]]] = {}
    ) -> str:
        completed_parameters = parameters.copy()
        completed_parameters.update({
            'action': action,
            'p_cert_id': self.cert.cert_id,
            'p_cert_key': self.cert.cert_key,
            'p_returntype': 'json',
        })
        parsed_url = urllib.parse.urlparse(self.endpoint)._replace(
            query=urllib.parse.urlencode(completed_parameters, doseq=True)
        )
        return urllib.parse.urlunparse(parsed_url)

    def request(self, action_endpoint: str) -> Dict[str, Any]:
        response = get(action_endpoint)
        response.raise_for_status()
        return response.json()

    def daily_sales_list(self) -> DailySalesListResponse:
        endpoint = self.build_action_endpoint('dailySalesList')
        response = self.request(endpoint)
        if response['error_code'] == ResultCode.SUCCESS.value:
            parsed_price = []
            for price in response['price']:
                parsed_price.append(DailySalesListPriceResponse(**price))
            response['price'] = parsed_price
        return DailySalesListResponse(**response)

    def period_product_list(
        self, params: PeriodProductListParameters
    ) -> PeriodProductListResponse:
        endpoint = self.build_action_endpoint(
            'periodProductList',
            {k: v for k, v in asdict(params).items() if v is not None}
        )
        response = self.request(endpoint)
        data = response.get('data')
        if data and isinstance(data, dict):
            parsed_item = []
            for item in data['item']:
                parsed_item.append(PeriodProductListItemResponse(**item))
            data['item'] = parsed_item
            response['data'] = PeriodProductListDataResponse(**data)
        return PeriodProductListResponse(**response)



""":mod:`kamis.models.daily_sales_list` --- DailySalesList API 모델
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KAMIS 오픈 API 중 DailySalesList API의 응답에 대응하는 모델을 정의한
모듈입니다.

https://www.kamis.or.kr/customer/reference/openapi_list.do?action=detail&boardno=6

"""
import collections
from dataclasses import InitVar, dataclass, field
import datetime
from typing import List, Optional, OrderedDict as OrderedDictType, Union

from .common import BasicResponse, Direction, ProductClass, ResultCode


__all__ = 'PriceResponse', 'Response'


@dataclass(frozen=True)
class PriceResponse:
    #: API 문서의 "구분(01:소매, 02:도매)"
    product_cls_code: InitVar[str]
    #: API 문서의 "구분 이름"; 사용되지 않습니다.
    product_cls_name: InitVar[str]
    #: 품목 구분; `product_cls_code` 값을 통해 만들어집니다.
    product_class: ProductClass = field(init=False)
    #: API 문서의 "부류코드"
    category_code: str
    #: API 문서의 "부류명"
    category_name: str
    #: API 문서의 "품목코드"
    productno: str
    #: API 문서의 "최근조사일"
    lastest_day: InitVar[str]
    #: 최근 조사일; `lastest_day` 값을 통해 만들어집니다.
    latest_date: datetime.date = field(init=False)
    #: API 문서의 "품목명"
    productName: InitVar[str]
    #: 품목명; `productName` 값을 통해 만들어집니다.
    product_name: str = field(init=False)
    #: API 문서의 "품목명"
    item_name: str
    #: API 문서의 "단위"
    unit: str
    #: API 문서의 "최근 조사일자"
    day1: InitVar[str]
    #: API 문서의 "최근 조사일자 가격"
    dpr1: InitVar[Union[List, str]]
    #: API 문서의 "1일전 일자"
    day2: InitVar[str]
    #: API 문서의 "1일전 가격"
    dpr2: InitVar[Union[List, str]]
    #: API 문서의 "1개월전 일자"
    day3: InitVar[str]
    #: API 문서의 "1개월전 가격"
    dpr3: InitVar[Union[List, str]]
    #: API 문서의 "1년전 일자"
    day4: InitVar[str]
    #: API 문서의 "1년전 가격"
    dpr4: InitVar[Union[List, str]]
    #: 일자별 가격; `day[1-4]` 및 `dpr[1-4]` 값을 통해 만들어집니다.
    date_price_dict: OrderedDictType[str, Optional[int]] = field(init=False)
    #: API 문서의 "등락여부( 0:가격하락 1:가격상승 2:등락없음 )"
    direction: InitVar[str]
    #: 등락여부; `direction` 값을 통해 만들어집니다.
    direction_type: Direction = field(init=False)
    #: API 문서의 "등락율"
    value: InitVar[str]
    #: 등락율; `value` 값을 통해 만들어집니다.
    direction_value: float = field(init=False)

    def __post_init__(
        self, product_cls_code: str, product_cls_name: str, lastest_day: str,
        productName: str, day1: str, dpr1: Union[List, str], day2: str,
        dpr2: Union[List, str], day3: str, dpr3: Union[List, str], day4: str,
        dpr4: Union[List, str], direction: str, value: str
    ):
        object.__setattr__(
            self, 'product_class', ProductClass(product_cls_code)
        )
        object.__setattr__(
            self, 'latest_date', datetime.date.fromisoformat(lastest_day)
        )
        object.__setattr__(self, 'product_name', productName)
        date_price_dict = collections.OrderedDict()
        date_price_dict[day1] = int(dpr1.replace(',', '')) if dpr1 else None
        date_price_dict[day2] = int(dpr2.replace(',', '')) if dpr2 else None
        date_price_dict[day3] = int(dpr3.replace(',', '')) if dpr3 else None
        date_price_dict[day4] = int(dpr4.replace(',', '')) if dpr4 else None
        object.__setattr__(self, 'date_price_dict', date_price_dict)
        object.__setattr__(self, 'direction_type', Direction(direction))
        object.__setattr__(self, 'direction_value', float(value))


@dataclass(frozen=True)
class Response(BasicResponse):
    #: API 문서의 "응답 메세지"
    price: InitVar[Union[List[PriceResponse], str]]
    #: 응답 메세지; `price` 값을 통해 만들어집니다.
    prices: Optional[List[PriceResponse]] = field(init=False)
    #: API 문서의 "결과코드"; 문서상으로는 result_code로 표기되어 있습니다.
    error_code: InitVar[str]
    #: 결과 코드; `error_code` 값을 통해 만들어집니다.
    result_code: ResultCode = field(init=False)

    def __post_init__(self, price: PriceResponse, error_code: str):
        object.__setattr__(
            self, 'prices', price if isinstance(price, list) else None
        )
        object.__setattr__(self, 'result_code', ResultCode(error_code))



""":mod:`kamis.models.period_product_list` --- PeriodProductList API 모델
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

KAMIS 오픈 API 중 PeriodProductList API의 파라미터 및 응답에 대응하는 모델을
정의한 모듈입니다.

https://www.kamis.or.kr/customer/reference/openapi_list.do?action=detail&boardno=2

"""
from dataclasses import InitVar, dataclass, field
import datetime
from typing import List, Optional, Tuple, Union

from .common import BasicResponse, Country, ProductClass, ResultCode


__all__ = 'DataResponse', 'ItemResponse', 'Parameters', 'Response'


@dataclass
class Parameters:
    """PeriodProductList API 요청에 필요한 파라미터를 담는 클래스입니다."""
    #: 조회 기간
    date_range: InitVar[Tuple[datetime.date, datetime.date]]
    #: API 문서의 요청 변수 중 "조회기간(start date)"
    #: `date_range` 값을 통해 만들어집니다.
    p_startday: str = field(init=False)
    #: API 문서의 요청 변수 중 "조회기간(end date)"
    #: `date_range` 값을 통해 만들어집니다.
    p_endday: str = field(init=False)
    #: 품목 구분
    product_class: InitVar[Optional[ProductClass]]
    #: API 문서의 요청 변수 중 "구분 ( 01:소매, 02:도매, default:02 )"
    #: `product_class` 값을 통해 만들어집니다.
    p_productclscode: Optional[str] = field(init=False, default=None)
    #: 부류 코드
    item_category_code: InitVar[str]
    #: API 문서의 요청 변수 중 "부류코드"
    #: `item_category_code` 값을 통해 만들어집니다.
    p_itemcategorycode: str = field(init=False)
    #: 품목 코드
    item_code: InitVar[str]
    #: API 문서의 요청 변수 중 "품목코드"; `item_code` 값을 통해 만들어집니다.
    p_itemcode: str = field(init=False)
    #: 품종 코드
    kind_code: InitVar[str]
    #: API 문서의 요청 변수 중 "품종코드"; `kind_code` 값을 통해 만들어집니다.
    p_kindcode: str = field(init=False)
    #: 등급 코드
    product_rank_code: InitVar[str]
    #: API 문서의 요청 변수 중 "등급코드"
    #: `product_rank_code` 값을 통해 만들어집니다.
    p_productrankcode: str = field(init=False)
    #: 지역
    county: InitVar[Optional[Country]]
    #: API 문서의 요청 변수 중 "지역"; `county` 값을 통해 만들어집니다.
    #: 이하 문서상 설명:
    #:  소매가격 선택가능 지역 (
    #:    1101:서울, 2100:부산, 2200:대구, 2300:인천, 2401:광주,
    #:    2501:대전, 2601:울산, 3111:수원, 3211:춘천, 3311:청주, 3511:전주,
    #:    3711:포항, 3911:제주, 3113:의정부, 3613:순천, 3714:안동, 3814:창원,
    #:    3145:용인
    #:  )
    #:  도매가격 선택가능 지역 (
    #:    1101:서울, 2100:부산, 2200:대구, 2401:광주, 2501:대전
    #:  )
    #:  default : 전체지역
    p_countycode: Optional[str] = field(init=False, default=None)
    #: kg 단위 환산 여부
    convert_kg: InitVar[Optional[bool]]
    #: API 문서의 요청 변수 중 "kg단위 환산여부"
    #: `convert_kg` 값을 통해 만들어집니다.
    #: 이하 문서상 설명:
    #:   Y : 1kg 단위표시, N : 정보조사 단위표시, ex: 쌀 20kg
    #:   default : N
    p_convert_kg_yn: Optional[str] = field(init=False, default=None)

    def __post_init__(
        self, date_range: Tuple[datetime.date, datetime.date],
        product_class: Optional[ProductClass], item_category_code: str,
        item_code: str, kind_code: str, product_rank_code: str,
        county: Optional[Country], convert_kg: Optional[bool]
    ):
        self.p_startday = date_range[0].isoformat()
        self.p_endday = date_range[1].isoformat()
        if product_class:
            self.p_productclscode = product_class.value
        self.p_itemcategorycode = item_category_code
        self.p_itemcode = item_code
        self.p_kindcode = kind_code
        self.p_productrankcode = product_rank_code
        if county:
            self.p_countycode = county.value
        if convert_kg:
            self.p_convert_kg_yn = 'Y' if convert_kg else 'N'


@dataclass(frozen=True)
class ItemResponse:
    #: API 문서의 "품목명"
    itemname: InitVar[Union[str, List]]
    #: "품목명"; `itemname` 값을 통해 만들어집니다.
    item_name: Optional[str] = field(init=False)
    #: API 문서의 "품종명"
    kindname: InitVar[Union[str, List]]
    #: "품종명"; `kindname` 값을 통해 만들어집니다.
    kind_name: Optional[str] = field(init=False)
    #: API 문서의 "시군구"
    countyname: InitVar[str]
    #: "시군구"; `countyname` 값을 통해 만들어집니다.
    county_name: str = field(init=False)
    #: API 문서의 "마켓명"
    marketname: InitVar[Union[str, List]]
    #: "마켓명"; `marketname` 값을 통해 만들어집니다.
    market_name: Optional[str] = field(init=False)
    #: API 문서의 "연도"
    yyyy: InitVar[str]
    #: API 문서의 "날짜"
    regday: InitVar[str]
    #: 조사일; `yyyy` 및 `regday` 값을 통해 만들어집니다.
    reg_date: datetime.date = field(init=False)
    #: API 문서의 "가격"
    price: InitVar[str]
    #: "가격"; `price` 값을 통해 만들어집니다.
    price_: Optional[int] = field(init=False)

    def __post_init__(
        self, itemname: Union[str, List], kindname: Union[str, List],
        countyname: str, marketname: Union[str, List], yyyy: str, regday: str,
        price: str
    ):
        object.__setattr__(self, 'item_name', itemname if itemname else None)
        object.__setattr__(self, 'kind_name', kindname if kindname else None)
        object.__setattr__(self, 'county_name', countyname)
        object.__setattr__(
            self, 'market_name', marketname if marketname else None
        )
        object.__setattr__(
            self,
            'reg_date',
            datetime.date.fromisoformat(f'{yyyy}-{regday.replace("/", "-")}')
        )
        object.__setattr__(
            self,
            'price_',
            int(price.replace(',', '')) if price != '-' else None
        )


@dataclass(frozen=True)
class DataResponse:
    #: 품목 가격 정보
    item: InitVar[List[ItemResponse]]
    #: 품목 가격 정보; `item` 값을 통해 만들어집니다.
    items: List[ItemResponse] = field(init=False)
    #: API 문서의 "결과코드"; 문서상으로는 result_code로 표기되어 있습니다.
    error_code: InitVar[str]
    #: 결과 코드; `error_code` 값을 통해 만들어집니다.
    result_code: ResultCode = field(init=False)

    def __post_init__(self, item: List[ItemResponse], error_code: str):
        object.__setattr__(self, 'items', item)
        object.__setattr__(self, 'result_code', ResultCode(error_code))


@dataclass(frozen=True)
class Response(BasicResponse):
    #: API 결과 응답 본문
    data: InitVar[Union[DataResponse, List[str]]]
    #: API 결과; `data` 값을 통해 만들어집니다.
    data_: DataResponse = field(init=False)

    def __post_init__(self, data: Union[DataResponse, List[str]]):
        object.__setattr__(
            self,
            'data_',
            data if isinstance(data, DataResponse) else DataResponse(
                item=[], error_code=data[0]
            )
        )



import pathlib

from setuptools import find_packages, setup

from kamis import __version__


__all__ = 'readme', 'setup'


def readme() -> str:
    name = pathlib.Path('README.rst')
    try:
        with open(name.resolve(), encoding='utf-8') as f:
            return f.read()
    except (IOError, OSError):
        return


setup(
    name='kamispy',
    version=__version__,
    description='KAMIS Open-API 비공식 파이썬 클라이언트',
    long_description=readme(),
    long_description_content_type='text/x-rst',
    url='https://github.com/spoqa/kamispy',
    license='MIT License',
    author='Spoqa Creators',
    author_emai='dev' '@' 'spoqa.com',
    packages=find_packages(exclude=('tests', 'tests.*')),
    python_requires='>=3.7.0',
    install_requires=['requests'],
    extras_require=dict(
        tests=['flake8', 'pytest', 'requests_mock']
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ]
)
